from typing import Any
import warnings

import dash


def _get_deprecated_webviz_settings(self: Any) -> dict:
    warnings.warn(
        "Accessing webviz_settings through the Dash application object has been deprecated, "
        "see https://github.com/equinor/webviz-config/pull/368",
        DeprecationWarning,
        stacklevel=2,
    )
    # pylint: disable=protected-access
    return self._deprecated_webviz_settings


def deprecate_webviz_settings_attribute_in_dash_app() -> None:
    """Helper that monkey patches dash.Dash application class so that access to
    the webviz_settings via the Dash application instance attribute is reported
    as being deprecated.
    """
    dash.Dash.webviz_settings = property(
        _get_deprecated_webviz_settings,
        None,
        None,
        "Property to mark webviz_settings access as deprecated",
    )



"""Builds automatic documentation of the installed webviz-config plugins.
The documentation is designed to be used by the YAML configuration file end
user. Sphinx has not been used, as the documentation from Sphinx is geared
mostly towards Python end users. It is also a small task generating `webviz`
documentation, and most of the Sphinx machinery is not needed.

Overall workflow is:
    * Find all installed plugins.
    * Automatically read docstring and __init__ function signatures (both
      argument names and which arguments have default values).
    * Output the extracted plugin information into docsify input using jinja2.
"""

import shutil
import inspect
import pathlib
from importlib import import_module
from collections import defaultdict
from typing import Any, Dict, Optional, Tuple, List, TypedDict

import jinja2

import webviz_config.plugins
from webviz_config.plugins import PLUGIN_METADATA, PLUGIN_PROJECT_METADATA
from .._config_parser import SPECIAL_ARGS
from .. import _deprecation_store as _ds


class ArgInfo(TypedDict, total=False):
    required: bool
    default: Any
    typehint: Any
    typehint_string: str


class PluginInfo(TypedDict):
    arg_info: Dict[str, ArgInfo]
    argument_description: Optional[str]
    data_input: Optional[str]
    description: Optional[str]
    dist_name: str
    dist_version: str
    name: str
    package_doc: Optional[str]


class ArgumentInfo(TypedDict, total=False):
    description: str
    default_value: str
    required: bool
    typehint: Any
    typehint_string: str
    deprecated: bool
    deprecation_message: str
    deprecation_code: str


def _find_plugin_deprecated_arguments(plugin: Any) -> Dict[str, Tuple[str, str]]:
    deprecated_arguments = (
        _ds.DEPRECATION_STORE.get_stored_plugin_argument_deprecations(plugin.__init__)
    )
    result: Dict[str, Tuple[str, str]] = {}
    for deprecated_argument in deprecated_arguments:
        if isinstance(deprecated_argument, _ds.DeprecatedArgumentCheck):
            for arg in deprecated_argument.argument_names:
                result[arg] = ("", deprecated_argument.callback_code)
    # Completely deprecated arguments have priority over deprecation check functions
    for deprecated_argument in deprecated_arguments:
        if isinstance(deprecated_argument, _ds.DeprecatedArgument):
            result[deprecated_argument.argument_name] = (
                deprecated_argument.long_message,
                "",
            )
    return result


def _extract_init_arguments_and_check_for_deprecation(
    reference: Any,
) -> Tuple[bool, Dict[str, ArgumentInfo], str]:
    """Returns all arguments of the given class' __init__ function including
    related documentation strings, typehints and deprecation information.
    """
    result: Dict[str, ArgumentInfo] = {}
    deprecated_arguments = _find_plugin_deprecated_arguments(reference)
    deprecation_check_code = ""

    argspec = inspect.getfullargspec(reference.__init__)
    result = {arg: ArgumentInfo() for arg in argspec.args if arg not in SPECIAL_ARGS}

    if argspec.defaults is not None:
        for arg, default in dict(
            zip(reversed(argspec.args), reversed(argspec.defaults))
        ).items():
            # Casting pathlib.Path to str could become unnecessary if outsourcing creation
            # of json schema to pydantic https://pydantic-docs.helpmanual.io/
            result[arg]["default_value"] = (
                str(default) if isinstance(default, pathlib.Path) else default
            )

    for arg, arg_info in result.items():
        arg_info["required"] = "default_value" not in arg_info

    for arg, annotation in argspec.annotations.items():
        if arg not in SPECIAL_ARGS and arg != "return":
            result[arg]["typehint"] = annotation
            result[arg]["typehint_string"] = _annotation_to_string(annotation)

    for arg, arg_info in result.items():
        if arg in deprecated_arguments:
            arg_info["deprecated"] = True
            arg_info["deprecation_message"] = deprecated_arguments[arg][0]
            arg_info["deprecation_code"] = deprecated_arguments[arg][1]
            if deprecation_check_code == "" and deprecated_arguments[arg][1] != "":
                deprecation_check_code = deprecated_arguments[arg][1]
        else:
            arg_info["deprecated"] = False

    return (bool(deprecated_arguments), result, deprecation_check_code)


def _document_plugin(plugin_name: str) -> PluginInfo:
    """Takes in plugin name as string and returns
    a dictionary according to the type definition PluginInfo.
    """

    reference = webviz_config.plugins.__getattr__(plugin_name)
    docstring = reference.__doc__ if reference.__doc__ is not None else ""
    docstring_parts = _split_docstring(docstring)
    module = inspect.getmodule(reference)
    subpackage = inspect.getmodule(module).__package__  # type: ignore
    dist_name = PLUGIN_METADATA[plugin_name]["dist_name"]
    (
        has_deprecated_arguments,
        arguments,
        deprecation_check_code,
    ) = _extract_init_arguments_and_check_for_deprecation(reference)
    deprecated = _ds.DEPRECATION_STORE.get_stored_plugin_deprecation(reference)

    plugin_info: PluginInfo = {
        "arg_info": arguments,
        "argument_description": docstring_parts[1]
        if len(docstring_parts) > 1
        else None,
        "data_input": docstring_parts[2] if len(docstring_parts) > 2 else None,
        "description": docstring_parts[0] if docstring != "" else None,
        "name": plugin_name,
        "package_doc": import_module(subpackage).__doc__,  # type: ignore
        "dist_name": dist_name,
        "dist_version": PLUGIN_PROJECT_METADATA[dist_name]["dist_version"],
        "deprecated": deprecated is not None,
        "deprecation_text_short": deprecated.short_message if deprecated else "",
        "deprecation_text_long": deprecated.long_message if deprecated else "",
        "has_deprecated_arguments": has_deprecated_arguments,
        "deprecation_check_code": deprecation_check_code,
    }

    return plugin_info


def get_plugin_documentation() -> defaultdict:
    """Find all installed webviz plugins, and then document them
    by grabbing docstring and input arguments / function signature.
    """

    plugin_doc = [
        _document_plugin(plugin)
        for plugin in webviz_config.plugins.__all__
        if not plugin.startswith("Example")
    ]

    # Sort the plugins by package:
    package_ordered: defaultdict = defaultdict(lambda: {"plugins": []})
    for sorted_plugin in sorted(plugin_doc, key=lambda x: (x["dist_name"], x["name"])):
        package = sorted_plugin["dist_name"]
        package_ordered[package]["plugins"].append(sorted_plugin)
        package_ordered[package]["doc"] = sorted_plugin["package_doc"]
        package_ordered[package]["dist_version"] = sorted_plugin["dist_version"]

    return package_ordered


def _annotation_to_string(annotation: Any) -> str:
    """Takes in a type annotation (that could come from e.g. inspect.getfullargspec)
    and transforms it into a human readable string.
    """

    def remove_fix(string: str, fix: str, prefix: bool = True) -> str:
        if prefix and string.startswith(fix):
            return string[len(fix) :]
        if not prefix and string.endswith(fix):
            return string[: -len(fix)]
        return string

    text_type = str(annotation)
    text_type = remove_fix(text_type, "typing.")
    text_type = remove_fix(text_type, "<class '")
    text_type = remove_fix(text_type, "'>", prefix=False)
    text_type = text_type.replace("pathlib.Path", "str (corresponding to a path)")

    return text_type


def build_docs(build_directory: pathlib.Path) -> None:

    # From Python 3.8, copytree gets an argument dirs_exist_ok.
    # Then the rmtree command can be removed.
    shutil.rmtree(build_directory)
    shutil.copytree(
        pathlib.Path(__file__).resolve().parent / "static",
        build_directory,
    )

    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )

    plugin_documentation = get_plugin_documentation()

    has_deprecated_plugins = {}
    has_deprecated_arguments = {}
    for dist_name, package_doc in plugin_documentation.items():
        if any(plugin["deprecated"] for plugin in package_doc["plugins"]):
            has_deprecated_plugins[dist_name] = package_doc
        if any(plugin["has_deprecated_arguments"] for plugin in package_doc["plugins"]):
            has_deprecated_arguments[dist_name] = package_doc
    template = template_environment.get_template("README.md.jinja2")
    for dist_name, package_doc in plugin_documentation.items():
        (build_directory / f"{dist_name}.md").write_text(
            template.render({"dist_name": dist_name, "package_doc": package_doc})
        )

    template = template_environment.get_template("sidebar.md.jinja2")
    (build_directory / "sidebar.md").write_text(
        template.render(
            {
                "packages": plugin_documentation.keys(),
                "deprecated_plugins": bool(has_deprecated_plugins),
                "deprecated_arguments": bool(has_deprecated_arguments),
            }
        )
    )

    template = template_environment.get_template("webviz-doc.js.jinja2")
    (build_directory / "webviz-doc.js").write_text(
        template.render(
            {"paths": ["/"] + [f"/{dist_name}" for dist_name in plugin_documentation]}
        )
    )

    template = template_environment.get_template("plugin_deprecations.md.jinja2")
    (build_directory / "plugin_deprecations.md").write_text(
        template.render({"packages": has_deprecated_plugins})
    )
    template = template_environment.get_template("argument_deprecations.md.jinja2")
    (build_directory / "argument_deprecations.md").write_text(
        template.render({"packages": has_deprecated_arguments})
    )


def _split_docstring(docstring: str) -> List[str]:
    """Divides docstring by splitting on ---, also unindents
    first in case of indented docstrings (similar to this one)
    """
    lines = docstring.strip().split("\n")

    try:
        indent_spaces = min(
            [len(line) - len(line.lstrip()) for line in lines[1:] if line.strip() != ""]
        )
    except ValueError:  # In the case of no original newline (\n)
        indent_spaces = 0

    unindented_lines = [lines[0]] + [line[indent_spaces:] for line in lines[1:]]
    return "\n".join(unindented_lines).split("\n---\n")



import os
import socket
from typing import Optional


def get_available_port(preferred_port: Optional[int] = None) -> int:
    """Finds an available port for use in webviz on localhost. If a reload process,
    it will reuse the same port as found in the parent process by using an inherited
    environment variable.

    If preferred_port is given, ports in the range [preferred_port, preferred_port + 20)
    will be tried first, before an OS provided random port is used as fallback.
    """

    def is_available(port: int) -> bool:
        with socket.socket() as sock:
            try:
                sock.bind(("localhost", port))
                return True
            except OSError:
                return False

    if os.environ.get("WEBVIZ_PORT") is None:
        port = None

        if preferred_port is not None:
            for port_to_test in range(preferred_port, preferred_port + 20):
                if is_available(port_to_test):
                    port = port_to_test
                    break

        if port is None:
            with socket.socket() as sock:
                sock.bind(("localhost", 0))
                port = sock.getsockname()[1]

        os.environ["WEBVIZ_PORT"] = str(port)
        return port

    return int(os.environ.get("WEBVIZ_PORT"))  # type: ignore[arg-type]



from dash import html, Dash, Input, Output

from .. import WebvizPluginABC


class ExamplePlugin(WebvizPluginABC):
    def __init__(self, app: Dash, title: str):
        super().__init__()

        self.title = title

        self.set_callbacks(app)

    @property
    def layout(self) -> html.Div:
        return html.Div(
            [
                html.H1(self.title),
                html.Button(
                    id=self.uuid("submit-button"), n_clicks=0, children="Submit"
                ),
                html.Div(id=self.uuid("output-state")),
            ]
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("output-state"), "children"),
            [Input(self.uuid("submit-button"), "n_clicks")],
        )
        def _update_output(n_clicks: int) -> str:
            return f"Button has been pressed {n_clicks} times."



import sys
import json
import shutil
import logging
import argparse
import pathlib
import subprocess

from ._user_data_dir import user_data_dir
from ._user_preferences import set_user_preferences, get_user_preference


def main() -> None:  # pylint: disable=too-many-locals,too-many-statements

    parser = argparse.ArgumentParser(
        prog=("Creates a Webviz Dash app from a configuration setup")
    )

    subparsers = parser.add_subparsers(
        metavar="SUBCOMMAND",
        required=True,
        help="Below are the available subcommands listed. "
        "Type e.g. 'webviz build --help' "
        "to get help on one particular "
        "subcommand.",
    )

    # Add "build" argument parser:

    parser_build = subparsers.add_parser("build", help="Build a Webviz Dash App")

    parser_build.add_argument(
        "yaml_file", type=pathlib.Path, help="Path to YAML configuration file"
    )
    parser_build.add_argument(
        "--portable",
        type=pathlib.Path,
        default=None,
        metavar="OUTPUTFOLDER",
        help="A portable webviz instance will created "
        "and saved to the given folder.",
    )
    parser_build.add_argument(
        "--theme",
        type=str,
        default=get_user_preference("theme")
        if get_user_preference("theme") is not None
        else "default",
        help="Which installed theme to use.",
    )
    parser_build.add_argument(
        "--loglevel",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        type=str,
        help="Wanted level of logging output from webviz plugins. "
        "Selecting e.g. INFO will show all log events of level INFO or higher "
        "(WARNING, ERROR and CRITICAL). Default level is WARNING.",
    )
    parser_build.add_argument(
        "--debug",
        action="store_true",
        help="Enable underlying Dash debug tools "
        "(UI with callback graph and errors, use development JavaScript "
        "bundles, validate types/values given to Dash component props).",
    )
    parser_build.add_argument(
        "--logconfig",
        type=pathlib.Path,
        default=None,
        metavar="CONFIGFILE",
        help="Path to YAML file with logging configuration.",
    )

    def parser_build_function(args: argparse.Namespace) -> None:
        from ._build_webviz import (  # pylint: disable=import-outside-toplevel
            build_webviz,
        )

        build_webviz(args)

    parser_build.set_defaults(func=parser_build_function)

    # Add "certificate" parser:

    parser_cert = subparsers.add_parser(
        "certificate",
        help="Create a https certificate authority for webviz "
        "(validity limited to localhost only)",
    )

    parser_cert.add_argument(
        "--force",
        action="store_true",
        help="Overwrite webviz root https certificate if it already exists",
    )

    parser_cert.add_argument(
        "--auto-install",
        action="store_true",
        help="Automatically install the webviz certificate in "
        "your personal public key infrastructure",
    )

    parser_cert.set_defaults(func=_dummy_create_ca)

    # Add "deploy" parser:

    parser_deploy = subparsers.add_parser(
        "deploy",
        help="Facilitates deployment of a portable Webviz application",
    )

    deployment_subparsers = parser_deploy.add_subparsers(
        metavar="DEPLOYMENT_METHOD",
        required=True,
        help="Below are the available deployment subcommands listed. "
        "Type e.g. 'webviz deploy radix --help' "
        "to get help on one particular "
        "subcommand.",
    )

    parser_radix_deploy = deployment_subparsers.add_parser(
        "radix",
        help="Deploy the app using Radix.",
    )

    parser_radix_deploy.add_argument(
        "portable_app",
        type=pathlib.Path,
        help="Path to portable Webviz application folder",
    )

    parser_radix_deploy.add_argument(
        "github_slug",
        type=str,
        help="GitHub slug (i.e. the string 'owner/repo_name').",
    )

    parser_radix_deploy.add_argument(
        "--initial-deploy",
        action="store_true",
        help="Add this flag when setting up a new up application for the first time. "
        "Do not include this if only overwriting an existing application.",
    )

    def parser_radix_deploy_function(args: argparse.Namespace) -> None:
        from ._deployment import (  # pylint: disable=import-outside-toplevel
            main_radix_deployment,
        )

        main_radix_deployment(args)

    parser_radix_deploy.set_defaults(func=parser_radix_deploy_function)

    # Add "documentation" parser:

    parser_docs = subparsers.add_parser(
        "docs",
        help="Get documentation on installed Webviz plugins",
    )

    parser_docs.add_argument(
        "--portable",
        type=pathlib.Path,
        default=None,
        metavar="OUTPUTFOLDER",
        help="Build documentation in given folder, "
        "which then can be deployed directly to e.g. GitHub pages.",
    )

    parser_docs.add_argument(
        "--force",
        action="store_true",
        help="Overwrite existing output (this flag "
        "only has effect if --portable is given)",
    )

    parser_docs.add_argument(
        "--skip-open",
        action="store_true",
        help="Skip opening the documentation automatically in browser.",
    )

    def parser_docs_function(args: argparse.Namespace) -> None:
        from ._docs.open_docs import (  # pylint: disable=import-outside-toplevel
            open_docs,
        )

        open_docs(args)

    parser_docs.set_defaults(func=parser_docs_function)

    # Add "preferences" parser:

    parser_preferences = subparsers.add_parser(
        "preferences",
        help="Set preferred webviz settings",
    )

    parser_preferences.add_argument(
        "--browser",
        type=str,
        help="Set the name of your preferred browser, "
        "in which localhost applications will open automatically.",
    )
    parser_preferences.add_argument(
        "--theme",
        type=str,
        help="Set your preferred Webviz theme, which will be used if "
        "'--theme' is not provided as an argument with the 'webviz build' command.",
    )

    def entrypoint_preferences(args: argparse.Namespace) -> None:

        if args.theme is not None:
            set_user_preferences(theme=args.theme)

        if args.browser is not None:
            set_user_preferences(browser=args.browser)

        print(f"Preferred theme: {get_user_preference('theme')}")
        print(f"Preferred browser: {get_user_preference('browser')}")

    parser_preferences.set_defaults(func=entrypoint_preferences)

    # Add "schema" parser:

    parser_schema = subparsers.add_parser(
        "schema",
        help="Create YAML (JSON) schema for webviz configuration "
        "file (including all installed plugins)",
    )

    parser_schema.add_argument(
        "--output",
        type=pathlib.Path,
        default=user_data_dir() / "webviz_schema.json",
        help="Name of output JSON schema file. If not given, "
        "it will be stored in your Webviz application settings folder.",
    )

    def parser_schema_function(args: argparse.Namespace) -> None:
        from ._docs._create_schema import (  # pylint: disable=import-outside-toplevel
            create_schema,
        )

        args.output.write_text(json.dumps(create_schema(), indent=4))
        print(f"Schema written to {args.output}")

    parser_schema.set_defaults(func=parser_schema_function)

    # Add "editor" parser:

    parser_editor = subparsers.add_parser(
        "editor",
        help="Create and edit Webviz configuration files",
    )

    # parser_editor.add_argument(
    #     "--path",
    #     type=pathlib.Path,
    #     help="Path to already existing Webviz configuration file.",
    # )

    def entrypoint_editor(  # pylint: disable=unused-argument
        args: argparse.Namespace,
    ) -> None:

        if sys.version_info < (3, 8):
            raise RuntimeError("Webviz editor requires at least Python 3.8")

        path_wce_executable = shutil.which("webviz-config-editor")
        if path_wce_executable is None:
            raise RuntimeError(
                "webviz-config-editor executable not found. You can download this from "
                "release assets (https://github.com/equinor/webviz-config-editor/releases)."
            )

        logging.warning(
            "Note that Webviz editor is in beta and early testing. "
            "Problems/bugs likely to occur."
        )

        command = [path_wce_executable]  # + ([] if args.path is None else [args.path])
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError:
            subprocess.run(command + ["--no-sandbox"], check=True)

    parser_editor.set_defaults(func=entrypoint_editor)

    # Do the argument parsing:

    args = parser.parse_args()

    args.func(args)


def _dummy_create_ca(_args: argparse.Namespace) -> None:
    """
    Print out a message about certs being unnecessary and exit gracefully (ie.
    returncode 0)
    """
    print(
        "The 'certificate' command is no longer needed as Webviz uses HTTP for local servers"
    )
    sys.exit(0)



from typing import Any

import flask


def silence_flask_startup() -> None:
    # pylint: disable=line-too-long
    """Calling this function monkey patches the function flask.cli.show_server_banner
    (https://github.com/pallets/flask/blob/a3f07829ca03bf312b12b3732e917498299fa82d/src/flask/cli.py#L657-L683)
    which by default outputs something like:

     * Serving Flask app "webviz_app" (lazy loading)
     * Environment: production
       WARNING: This is a development server. Do not use it in a production deployment.
       Use a production WSGI server instead.
     * Debug mode: off

    This warning is confusing to new users of flask and webviz-config, which thinks
    something is wrong (even though having development/debug mode turned off, and
    limit availability to localhost, is best practice wrt. security).

    After calling this function the exact lines above are not shown
    (all other information/output from the flask instance is untouched).
    """

    def silent_function(*_args: Any, **_kwargs: Any) -> None:
        pass

    flask.cli.show_server_banner = silent_function



class WebvizFactory:
    """Base class for all factories that want to register themselves in
    WebvizFactoryRegistry.

    Note that this functionality is provisional/experimental, and will not necessarily
    see a deprecation phase and "may deviate from the usual version semantics."
    """

    def cleanup_resources_after_plugin_init(self) -> None:
        """Will be called after all plugins have been initialized to allow the factory
        to do clean-up of any temporary resources allocated during the initialization
        phase. The base implementation does nothing, override this function to
        perform factory-specific cleanup.
        """



import warnings
from unittest import mock
import importlib
import importlib.metadata

import webviz_config.plugins._utils


class DistMock:
    # pylint: disable=too-few-public-methods
    def __init__(self, entry_points, name):
        self.metadata = {"name": name}

        self.entry_points = entry_points
        self.version = "123"


plugin_entrypoint_mock1 = mock.Mock()
plugin_entrypoint_mock1.group = "webviz_config_plugins"
plugin_entrypoint_mock1.name = "SomePlugin1"

plugin_entrypoint_mock2 = mock.Mock()
plugin_entrypoint_mock2.group = "webviz_config_plugins"
plugin_entrypoint_mock2.name = "SomePlugin2"

dist_mock1 = DistMock([plugin_entrypoint_mock1], "dist_mock1")
dist_mock2 = DistMock([plugin_entrypoint_mock1], "dist_mock2")
dist_mock3 = DistMock([plugin_entrypoint_mock2], "dist_mock3")


def test_no_warning(monkeypatch):
    # pylint: disable=protected-access
    monkeypatch.setattr(importlib.metadata, "requires", lambda x: [])
    importlib.reload(webviz_config.plugins._utils)

    with warnings.catch_warnings(record=True) as warn:
        (
            metadata,
            _,
            plugin_entrypoints,
        ) = webviz_config.plugins._utils.load_webviz_plugins_with_metadata(
            [dist_mock1, dist_mock3]
        )
        assert len(warn) == 0, "Too many warnings"

    assert len(metadata) == 2, "Wrong number of items in metadata"
    assert "SomePlugin1" in plugin_entrypoints
    assert "SomePlugin2" in plugin_entrypoints


def test_warning_multiple(monkeypatch):
    # pylint: disable=protected-access
    monkeypatch.setattr(importlib.metadata, "requires", lambda x: [])
    importlib.reload(webviz_config.plugins._utils)

    with warnings.catch_warnings(record=True) as warn:
        (
            metadata,
            _,
            plugin_entrypoints,
        ) = webviz_config.plugins._utils.load_webviz_plugins_with_metadata(
            [dist_mock1, dist_mock2]
        )

        assert len(warn) == 1
        assert issubclass(warn[-1].category, RuntimeWarning)
        print(warn[-1].message)
        assert str(warn[-1].message) == (
            "Multiple versions of plugin with name SomePlugin1. "
            "Already loaded from project dist_mock1. "
            "Overwriting using plugin with from project dist_mock2"
        )

    assert len(metadata) == 1, "Wrong number of items in metadata"
    assert metadata["SomePlugin1"]["dist_name"] == "dist_mock2", "Wrong dist name"
    assert "SomePlugin1" in plugin_entrypoints



from typing import Callable, List, Optional, Union
import abc

from ._layout_unique_id import LayoutUniqueId


class LayoutBaseABC(abc.ABC):
    def __init__(self) -> None:
        self._unique_id = LayoutUniqueId()
        self._plugin_register_id_func: Optional[
            Callable[[Union[str, List[str]]], None]
        ] = None
        self._plugin_get_store_unique_id_func: Optional[Callable[[str], str]] = None

    def _set_plugin_register_id_func(
        self, func: Callable[[Union[str, List[str]]], None]
    ) -> None:
        self._plugin_register_id_func = func

    def _set_unique_id(self, parent_unique_id: LayoutUniqueId) -> None:
        self._unique_id.adopt(parent_unique_id)

        if self._plugin_register_id_func:
            self._plugin_register_id_func(str(self._unique_id))

    def get_unique_id(self) -> LayoutUniqueId:
        return self._unique_id

    def _set_plugin_get_store_unique_id_func(self, func: Callable[[str], str]) -> None:
        self._plugin_get_store_unique_id_func = func

    def get_store_unique_id(self, store_id: str) -> str:
        if self._plugin_get_store_unique_id_func:
            return self._plugin_get_store_unique_id_func(store_id)

        raise NotImplementedError("This element has not been added to a plugin yet.")



# pylint: disable=line-too-long
from typing import Any, Callable, get_args, get_origin, _TypedDictMeta, TypeVar, Union  # type: ignore[attr-defined]
import inspect

T = TypeVar("T")


class ConversionError(Exception):
    pass


def _isinstance(arg: Any, annotation: Any) -> bool:
    # pylint: disable=too-many-return-statements, too-many-branches
    if annotation is type(None) or annotation is None:
        return arg is None

    if annotation is Any:
        return True

    if get_origin(annotation) is None:
        try:
            return isinstance(arg, annotation)
        except TypeError:
            return False

    if get_origin(annotation) == Union:
        for annotation_arg in get_args(annotation):
            if _isinstance(arg, annotation_arg):
                return True

    if get_origin(annotation) is list and isinstance(arg, list):
        result = True
        type_args = get_args(annotation)
        if len(type_args) == 1:
            for annotation_arg in arg:
                result &= _isinstance(annotation_arg, type_args[0])
        return result

    if get_origin(annotation) is dict and isinstance(arg, dict):
        result = True
        type_args = get_args(annotation)
        if len(type_args) == 2:
            for key, value in arg.items():
                result &= _isinstance(key, type_args[0])
                result &= _isinstance(value, type_args[1])
        return result

    return False


def convert(arg: Any, convert_to: T) -> T:
    # pylint: disable=too-many-return-statements, too-many-branches
    additional_error_message: str = ""
    try:
        if _isinstance(arg, convert_to):
            return arg
        if (
            inspect.isclass(convert_to)
            and not isinstance(convert_to, _TypedDictMeta)
            and arg is not None
        ):
            return convert_to(arg)
        if (
            isinstance(convert_to, _TypedDictMeta)
            and "__annotations__" in dir(convert_to)
            and isinstance(arg, dict)
        ):
            new_dict = convert_to()
            for key, value in arg.items():
                if key in list(convert_to.__annotations__.keys()):
                    new_dict[key] = convert(value, convert_to.__annotations__[key])
                else:
                    raise Exception(
                        f"""
                        Key '{key}' not allowed in '{convert_to}'.\n
                        Allowed keys are: {', '.join(list(convert_to.__annotations__.keys()))}
                        """
                    )

            if not convert_to.__total__ or len(new_dict.keys()) == len(
                convert_to.__annotations__.keys()
            ):
                return new_dict

        if convert_to is list and isinstance(arg, list):
            return arg  # type: ignore[return-value]
        if get_origin(convert_to) is list and isinstance(arg, list):
            return [convert(a, get_args(convert_to)[0]) for a in arg]  # type: ignore[return-value]
        if convert_to is dict and isinstance(arg, dict):
            return arg  # type: ignore[return-value]
        if get_origin(convert_to) is dict and isinstance(arg, dict):
            if len(get_args(convert_to)) == 2:
                return {  # type: ignore[return-value]
                    convert(key, get_args(convert_to)[0]): convert(
                        value, get_args(convert_to)[1]
                    )
                    for key, value in arg.items()
                }
        if get_origin(convert_to) is Union and "__args__" in dir(convert_to):
            for convert_type in get_args(convert_to):
                try:
                    return convert(arg, convert_type)
                except ConversionError:
                    pass

    # pylint: disable=broad-except
    except Exception as exception:
        additional_error_message = f"\n\nMore details:\n{exception}"

    raise ConversionError(
        f"Argument of type '{type(arg)}' cannot be converted to type '{convert_to}'.{additional_error_message}"
    )


def callback_typecheck(func: Callable) -> Callable:
    signature = inspect.signature(func)
    parameters = list(signature.parameters.values())

    def wrapper(*_args) -> signature.return_annotation:  # type: ignore[no-untyped-def,name-defined]
        adjusted_args: list = []

        for index, arg in enumerate(_args):
            try:
                adjusted_args.append(convert(arg, parameters[index].annotation))
            except ConversionError as exception:
                raise ConversionError(
                    f"Error while converting input to argument '{parameters[index].name}' of function '{func.__name__}' in file '{func.__globals__['__file__']}': {exception}"
                ) from exception

        return func(*adjusted_args)

    return wrapper



import os


def is_reload_process() -> bool:
    """Within the flask reload machinery, it is not straight forward to know
    if the code is run as the main process (i.e. the process the user directly
    started), or if the code is a "hot reload process" (see Flask
    documentation).

    This utility function will use the fact that the reload process
    sets an environment variable WERKZEUG_RUN_MAIN.
    """

    return os.environ.get("WERKZEUG_RUN_MAIN") == "true"



from ._settings_group_abc import SettingsGroupABC
from ._views import ViewABC, ViewElementABC, ViewLayoutElement, LayoutElementType
from ._layout_base_abc import LayoutBaseABC
from ._layout_unique_id import LayoutUniqueId



import copy
import json


class WebvizConfigTheme:
    """Webviz config themes are all instances of this class. The only mandatory
    property is the theme name set at initialization.
    """

    def __init__(self, theme_name: str):
        self.theme_name = theme_name

        self._csp = {
            "default-src": "'none'",
            "connect-src": "'self'",
            "prefetch-src": "'self'",
            "style-src": ["'self'", "'unsafe-inline'"],  # [1]
            "script-src": ["'self'", "blob:", "'unsafe-eval'"],  # [blob: 2] [eval: 3]
            "img-src": ["'self'", "data:", "blob:"],  # [4]
            "navigate-to": "'self'",
            "base-uri": "'self'",
            "form-action": "'self'",
            "frame-ancestors": "'self'",  # [5]
            "frame-src": "'self'",  # [5]
            "object-src": "'self'",
        }

        """
        These are the current exceptions to the most strict CSP setup:
            [1] unsafe-inline for style still needed by plotly
                (https://github.com/plotly/plotly.js/issues/2355)
            [2] Computational heavy Dash components might want to use web workers.
                Since it is not easy in Dash to host related worker scripts, these
                worker scripts are typically created on demand and given as blob: URLs.
                We do not allow data: script URLs since they do not have the same
                security feature as blob: URLs which can only be created by the browser
                itself and have scope limited to the window it was created in.
            [3] unsafe-eval still needed for plotly.js bundle
                (https://github.com/plotly/plotly.js/issues/897)
            [4] html2canvas in webviz-core-components needs "data:" in img-src to create
                screenshots, while plotly.js "Download screenshot to png" requires
                "blob:" in img-src.
            [5] We use 'self' instead of 'none' due to what looks like a Chromium bug,
                where e.g. pdf's included using <embed> is not rendered. Might be
                related to https://bugs.chromium.org/p/chromium/issues/detail?id=1002610
        """

        self._feature_policy = {
            "camera": "'none'",
            "geolocation": "'none'",
            "microphone": "'none'",
            "payment": "'none'",
        }

        self._external_stylesheets: list = []
        self._assets: list = []
        self._plotly_theme: dict = {}

    def to_json(self) -> str:
        return json.dumps(vars(self), indent=4, sort_keys=True)

    def from_json(self, json_string: str) -> None:
        for key, value in json.loads(json_string).items():
            setattr(self, key, value)

    def adjust_csp(self, dictionary: dict, append: bool = True) -> None:
        """If the default CSP settings needs to be changed, this function can
        be called by giving in a dictionary with key-value pairs which should
        be changed. If `append=True`, the CSP sources given in the dictionary
        are appended to the whitelisted sources in the default configuration.
        If not the input value source list is instead replacing the default
        whitelisted sources.
        """

        for key, value in dictionary.items():
            if append and key in self._csp:
                self._csp[key] += value
            else:
                self._csp[key] = value

    def create_themed_layout(self, layout: dict) -> dict:
        """
        Create a new Plotly layout dict by merging the input layout with the theme layout,
        prioritizing the input layout if there are conflicts with the theme layout. In addition:
        For the special case of multiple xaxes or yaxes, e.g. xaxis2 and xaxis3 (for a secondary
        and tertiary xaxis), the axis will get the theme xaxis/yaxis layout, unless they are
        defined themselves as e.g. xaxis2 in the theme layout. Note that e.g. xaxis2 still needs to
        be in the input layout, just not in the theme layout.
        """
        # pylint: disable=too-many-nested-blocks
        def deep_update(update: dict, ref: dict) -> dict:
            for key, value in ref.items():
                if key in update:
                    if isinstance(value, dict):
                        if isinstance(update[key], dict):
                            update[key] = deep_update(update[key], value)
                        if key in ["xaxis", "yaxis"]:
                            for kkey in update:
                                if kkey not in ref and kkey.startswith(key):
                                    update[kkey] = deep_update(update[kkey], value)
                else:
                    update[key] = value
            return update

        return deep_update(
            copy.deepcopy(layout), copy.deepcopy(self._plotly_theme["layout"])
        )

    @property
    def csp(self) -> dict:
        """Returns the content security policy settings for the theme."""
        return self._csp

    @property
    def feature_policy(self) -> dict:
        """Returns the feature policy settings for the theme."""
        return self._feature_policy

    @property
    def plotly_theme(self) -> dict:
        return copy.deepcopy(self._plotly_theme)

    @plotly_theme.setter
    def plotly_theme(self, plotly_theme: dict) -> None:
        """Layout object of Plotly graph objects."""
        self._plotly_theme = plotly_theme

    def plotly_theme_layout_update(self, plotly_theme_layout: dict) -> None:
        """Updates layout object of Plotly graph objects based on input dict"""
        self._plotly_theme["layout"] = self.create_themed_layout(plotly_theme_layout)

    @property
    def external_stylesheets(self) -> list:
        return self._external_stylesheets

    @external_stylesheets.setter
    def external_stylesheets(self, external_stylesheets: list) -> None:
        """Set optional external stylesheets to be used in the Dash
        application. The input variable `external_stylesheets` should be
        a list."""
        self._external_stylesheets = external_stylesheets

    @property
    def assets(self) -> list:
        return self._assets

    @assets.setter
    def assets(self, assets: list) -> None:
        """Set optional theme assets to be copied over to the `./assets` folder
        when the webviz dash application is created. The input variable
        `assets` should be a list of absolute file paths to the different
        assets.
        """
        self._assets = assets



import shutil
import pathlib
import tempfile
import argparse
import logging

import flask

import webviz_config.utils
from ._build_docs import build_docs


def _start_doc_app(build_directory: pathlib.Path) -> None:

    app = flask.Flask(__name__, static_folder=str(build_directory), static_url_path="")
    app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0

    @app.route("/")
    def _index() -> str:
        return (build_directory / "index.html").read_text()

    webviz_config.utils.silence_flask_startup()
    logging.getLogger("werkzeug").setLevel(logging.WARNING)

    port = webviz_config.utils.get_available_port(preferred_port=5050)
    token = webviz_config.LocalhostToken(app, port).one_time_token
    webviz_config.utils.LocalhostOpenBrowser(port, token)

    app.run(
        host="localhost",
        port=port,
        debug=False,
    )


def open_docs(args: argparse.Namespace) -> None:

    if args.portable is None:
        build_directory = pathlib.Path(tempfile.mkdtemp())
    else:
        build_directory = args.portable.resolve()
        if build_directory.exists():
            if not args.force:
                raise ValueError(
                    f"{build_directory} already exists. Either add --force or change output folder."
                )
            shutil.rmtree(build_directory)
        build_directory.mkdir(parents=True)

    try:
        build_docs(build_directory)
        if not args.skip_open:
            _start_doc_app(build_directory)
    finally:
        if args.portable is None:
            shutil.rmtree(build_directory)



from typing import List, Optional


class LayoutUniqueId:
    def __init__(
        self,
        plugin_uuid: Optional[str] = None,
        view_id: Optional[str] = None,
        view_element_id: Optional[str] = None,
        settings_group_id: Optional[str] = None,
        component_id: Optional[str] = None,
        other: Optional["LayoutUniqueId"] = None,
    ) -> None:
        self._plugin_uuid = plugin_uuid
        self._view_id = view_id
        self._settings_group_id = settings_group_id
        self._view_element_id = view_element_id
        self._component_id = component_id

        if other:
            self.adopt(other)

    def get_plugin_uuid(self) -> Optional[str]:
        return self._plugin_uuid

    def get_view_unique_id(self) -> str:
        ids: List[str] = []
        if not self._view_id:
            return ""

        if self._plugin_uuid:
            ids.append(self._plugin_uuid)
        if self._view_id:
            ids.append(self._view_id)

        return "-".join(ids)

    def get_view_id(self) -> Optional[str]:
        return self._view_id

    def get_view_element_unique_id(self) -> str:
        ids: List[str] = []
        if self._plugin_uuid:
            ids.append(self._plugin_uuid)
        if self._view_id:
            ids.append(self._view_id)
        if self._view_element_id:
            ids.append(self._view_element_id)

        return "-".join(ids)

    def get_view_element_id(self) -> Optional[str]:
        return self._view_element_id

    def get_component_id(self) -> Optional[str]:
        return self._component_id

    def get_settings_group_id(self) -> Optional[str]:
        return self._settings_group_id

    def get_settings_group_unique_id(self) -> str:
        ids: List[str] = []
        if self._plugin_uuid:
            ids.append(self._plugin_uuid)
        if self._view_id:
            ids.append(self._view_id)
        if self._view_element_id:
            ids.append(self._view_element_id)
        if self._settings_group_id:
            ids.append(self._settings_group_id)

        return "-".join(ids)

    def set_plugin_uuid(self, plugin_uuid: str) -> None:
        self._plugin_uuid = plugin_uuid

    def set_view_id(self, view_id: str) -> None:
        self._view_id = view_id

    def set_view_element_id(self, view_element_id: str) -> None:
        self._view_element_id = view_element_id

    def set_settings_group_id(self, settings_group_id: str) -> None:
        self._settings_group_id = settings_group_id

    def set_component_id(self, component_id: str) -> None:
        self._component_id = component_id

    def is_plugin(self) -> bool:
        return (
            self._plugin_uuid is not None
            and self._view_id is None
            and self._settings_group_id is None
        )

    def is_view(self) -> bool:
        return (
            self._view_id is not None
            and self._view_element_id is None
            and self._settings_group_id is None
        )

    def is_view_element(self) -> bool:
        return self._view_element_id is not None

    def is_settings_group(self) -> bool:
        return self._settings_group_id is not None

    def is_component(self) -> bool:
        return self._component_id is not None

    def adopt(self, other: "LayoutUniqueId") -> None:
        if self._plugin_uuid is None and other.get_plugin_uuid() is not None:
            self._plugin_uuid = other.get_plugin_uuid()

        if self._view_id is None and other.get_view_id() is not None:
            self._view_id = other.get_view_id()

        if self._view_element_id is None and other.get_view_element_id() is not None:
            self._view_element_id = other.get_view_element_id()

        if (
            self._settings_group_id is None
            and other.get_settings_group_id() is not None
        ):
            self._settings_group_id = other.get_settings_group_id()

        if self._component_id is None and other.get_component_id() is not None:
            self._component_id = other.get_component_id()

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        ids: List[str] = []
        if self._plugin_uuid:
            ids.append(self._plugin_uuid)
        if self._view_id:
            ids.append(self._view_id)
        if self._view_element_id:
            ids.append(self._view_element_id)
        if self._settings_group_id:
            ids.append(self._settings_group_id)
        if self._component_id:
            ids.append(self._component_id)

        return "-".join(ids)



import glob
import pathlib

from plotly.io import templates

from webviz_config import WebvizConfigTheme

default_theme = WebvizConfigTheme(theme_name="default")

default_theme.assets = glob.glob(
    str(pathlib.Path(__file__).resolve().parent / "default_assets" / "*")
)
default_theme.assets.append(
    str(
        pathlib.Path(__file__).resolve().parent.parent
        / "_docs"
        / "static"
        / "webviz-logo.svg"
    )
)

default_theme.plotly_theme = templates["plotly"].to_plotly_json()



from typing import Type, Union
from enum import Enum

from dash.development.base_component import Component
from dash import html

from webviz_config.webviz_plugin_subclasses import ViewElementABC


class TextViewElement(ViewElementABC):
    class Ids(str, Enum):
        TEXT = "text"

    def __init__(self) -> None:
        super().__init__()

    def inner_layout(self) -> Union[str, Type[Component]]:
        return html.Div(
            id=self.register_component_unique_id(TextViewElement.Ids.TEXT),
            children=[
                html.H1("Hello"),
                """
                This is an example plugin.
                Please have a look how views and settings are working in this new environment =).
                """,
            ],
        )



from typing import List

import pandas as pd

from .. import WebvizPluginABC
from ..webviz_store import webvizstore
from ..common_cache import CACHE


class ExamplePortable(WebvizPluginABC):
    def __init__(self, some_number: int):
        super().__init__()

        self.some_number = some_number
        self.some_string = "a"

    def add_webvizstore(self) -> List[tuple]:
        return [
            (
                input_data_function,
                [{"some_string": self.some_string, "some_number": self.some_number}],
            )
        ]

    @property
    def layout(self) -> str:
        return str(input_data_function(self.some_string, some_number=self.some_number))


@CACHE.memoize()
@webvizstore
def input_data_function(
    some_string: str, some_number: int, some_bool: bool = True
) -> pd.DataFrame:
    print("This time I'm actually doing the calculation...")
    return pd.DataFrame(
        data={
            "col1": [some_number, some_number * 2],
            "col2": [some_number * 3, some_number * 4],
            "col3": [some_string, some_string + "b"],
            "col4": [some_bool, not some_bool],
        }
    )



import os
import re
import pathlib

from setuptools import setup, find_packages


def get_long_description() -> str:
    """Converts relative repository links to absolute URLs
    if GITHUB_REPOSITORY and GITHUB_SHA environment variables exist.
    If not, it returns the raw content in README.md.
    """

    raw_readme = pathlib.Path("README.md").read_text()

    repository = os.environ.get("GITHUB_REPOSITORY")
    sha = os.environ.get("GITHUB_SHA")

    if repository is not None and sha is not None:
        full_url = f"https://github.com/{repository}/blob/{sha}/"
        return re.sub(r"]\((?!https)", "](" + full_url, raw_readme)
    return raw_readme


TESTS_REQUIRES = [
    "bandit",
    "black>=22.12",
    "dash[testing]",
    "jsonschema",
    "mypy",
    "pylint<=2.13.9",  # Locked due to https://github.com/equinor/webviz-subsurface/issues/1052
    "pytest-xdist",
    "pytest-forked",
    "selenium",
    "types-bleach",
    "types-markdown",
    "types-pyyaml",
    "types-requests",
]

# pylint: disable=line-too-long
setup(
    name="webviz-config",
    description="Configuration file support for webviz",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/webviz-config",
    author="R&T Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "webviz_config": [
            "_docs/static/*",
            "_docs/static/fonts/*",
            "py.typed",
            "static/*",
            "static/.dockerignore",
            "static/.gitignore",
            "static/assets/*",
            "templates/*",
            "themes/default_assets/*",
        ]
    },
    entry_points={
        "console_scripts": ["webviz=webviz_config.command_line:main"],
        "webviz_config_plugins": [
            "ExampleAssets = webviz_config.generic_plugins._example_assets:ExampleAssets",
            "ExampleWlfPlugin = webviz_config.generic_plugins._example_wlf_plugin:ExampleWlfPlugin",
            "ExampleDataDownload = webviz_config.generic_plugins._example_data_download:ExampleDataDownload",
            "ExamplePlugin = webviz_config.generic_plugins._example_plugin:ExamplePlugin",
            "ExamplePortable = webviz_config.generic_plugins._example_portable:ExamplePortable",
            "ExampleTour = webviz_config.generic_plugins._example_tour:ExampleTour",
            "BannerImage = webviz_config.generic_plugins._banner_image:BannerImage",
            "DataTable = webviz_config.generic_plugins._data_table:DataTable",
            "EmbedPdf = webviz_config.generic_plugins._embed_pdf:EmbedPdf",
            "Markdown = webviz_config.generic_plugins._markdown:Markdown",
            "SyntaxHighlighter = webviz_config.generic_plugins._syntax_highlighter:SyntaxHighlighter",
            "TablePlotter = webviz_config.generic_plugins._table_plotter:TablePlotter",
            "PivotTable = webviz_config.generic_plugins._pivot_table:PivotTable",
        ],
        "pytest11": ["webviz = webviz_config.testing._plugin"],
    },
    install_requires=[
        "bleach[css]>=5",
        "cryptography>=2.4",
        "dash>=2.0",
        "dash-pivottable>=0.0.2",
        "flask>=2.0",
        "flask-caching>=1.4",
        "flask-talisman>=0.6",
        "jinja2>=2.10",
        "markdown>=3.0",
        "msal>=1.5.0",
        "orjson>=3.3",
        "pandas>=1.0",
        "pyarrow>=0.16",
        "pyyaml>=5.1",
        "requests>=2.20",
        "tqdm>=4.8",
        "webviz-core-components>=0.6",
        "werkzeug>=2.0",
    ],
    extras_require={
        "tests": TESTS_REQUIRES,
        "deployment": [
            "aiohttp",  # https://github.com/Azure/azure-sdk-for-python/issues/19201
            "azure-core",
            "azure-identity",
            "azure-mgmt-resource",
            "azure-mgmt-storage",
            "azure-storage-blob",
        ],
    },
    setup_requires=["setuptools_scm~=3.2"],
    python_requires="~=3.8",
    use_scm_version=True,
    zip_safe=False,
    project_urls={
        "Documentation": "https://equinor.github.io/webviz-config",
        "Download": "https://pypi.org/project/webviz-config",
        "Source": "https://github.com/equinor/webviz-config",
        "Tracker": "https://github.com/equinor/webviz-config/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: MIT License",
    ],
)



from typing import Any, List, Optional, Dict, Callable, Union
from dataclasses import dataclass


@dataclass(frozen=True)
class DeprecatedPlugin:
    class_reference: Any
    short_message: str
    long_message: str


@dataclass(frozen=True)
class DeprecatedArgument:
    method_reference: Any
    method_name: str
    argument_name: str
    argument_value: str
    short_message: str
    long_message: str


@dataclass(frozen=True)
class DeprecatedArgumentCheck:
    method_reference: Any
    method_name: str
    argument_names: List[str]
    callback: Callable
    callback_code: str


class DeprecationStore:
    def __init__(self) -> None:
        self.stored_plugin_deprecations: Dict[Any, DeprecatedPlugin] = {}
        self.stored_plugin_argument_deprecations: List[
            Union[DeprecatedArgument, DeprecatedArgumentCheck]
        ] = []

    def register_deprecated_plugin(self, deprecated_plugin: DeprecatedPlugin) -> None:
        """This function is automatically called by the decorator
        @deprecated_plugin, registering the plugin it decorates.
        """
        self.stored_plugin_deprecations[
            deprecated_plugin.class_reference
        ] = deprecated_plugin

    def register_deprecated_plugin_argument(
        self,
        deprecated_plugin_argument: Union[DeprecatedArgument, DeprecatedArgumentCheck],
    ) -> None:
        """This function is automatically called by the decorator
        @deprecated_plugin_arguments, registering the __init__ function it decorates.
        """
        self.stored_plugin_argument_deprecations.append(deprecated_plugin_argument)

    def get_stored_plugin_deprecation(self, plugin: Any) -> Optional[DeprecatedPlugin]:
        return self.stored_plugin_deprecations.get(plugin)

    def get_stored_plugin_argument_deprecations(
        self, method: Callable
    ) -> List[Union[DeprecatedArgument, DeprecatedArgumentCheck]]:
        return [
            stored
            for stored in self.stored_plugin_argument_deprecations
            if stored.method_reference == method
        ]


DEPRECATION_STORE = DeprecationStore()



import pathlib
import pytest
import yaml
import jsonschema

from webviz_config._docs._create_schema import create_schema


@pytest.mark.parametrize(
    "config_file_path",
    [
        (pathlib.Path("examples") / "basic_example.yaml"),
        (pathlib.Path("examples") / "basic_example_advanced_menu.yaml"),
    ],
)
def test_schema(config_file_path: pathlib.Path):
    """Tests both that the generated schema is valid,
    and that the input configuration is valid according to the schema.
    """

    config = yaml.safe_load(config_file_path.read_text())
    jsonschema.validate(instance=config, schema=create_schema())
